////////////////////////////////////////////////////////////////////////
/// \file  SingleGen_plugin.cc
/// \brief Generator for cosmic-rays
///
/// Module designed to produce a set list of particles for a MC event
///
/// \version $Id: SingleGen.cxx,v 1.4 2010/03/29 09:54:01 brebel Exp $
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////




// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h" 
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "TFile.h"
#include "TTree.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TVector3.h"
#include "TDatabasePDG.h"
#include "TRandom3.h"
#include "TSystem.h"
#include "TSystemDirectory.h"
#include "TSystemFile.h"

#include <sstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <dirent.h>
#include <time.h>
#include <stdio.h>
#include <cstdlib>
#include <stdlib.h>
#include <math.h>
#include <vector>
#include <string>

// nutools includes
#include "SimulationBase/MCTruth.h"

// lar includes
#include "larcore/Geometry/Geometry.h"
#include "larcore/SummaryData/RunData.h"
#include "larcore/SummaryData/POTSummary.h"


namespace simb { class MCTruth; }

namespace t962 {

  /// module to produce single or multiple specified particles in the detector
  class TGMuon : public art::EDProducer {

  public:
    explicit TGMuon(fhicl::ParameterSet const& pset);
    virtual ~TGMuon();

    // This is called for each event.
    void produce(art::Event& evt);
    void beginRun(art::Run& run);
    void endSubRun(art::SubRun& sr);
    void reconfigure(fhicl::ParameterSet const& pset) {};

  private:

    void Sample(simb::MCTruth &mct);        
    
    TRandom3           *fRand;           ///< random number
    int                 fSeed;           ///< random number seed    

    char fBuffer[128];
    TDirectory* fSaveDir ;
    TString fpwd;
    int frun;
    int fsubRun;
    int fsnarl;
    int fntrkstp;
    float ftrkIndex;
    float ftrtgtd;
    float ftrkVtxX;
    float ftrkVtxY;
    float ftrkVtxZ;
    float ftrkdcosx;
    float ftrkdcosy;
    float ftrkdcosz; 
    float ftrkmom;
    float ftrkqp;
    float totalpot;

    int fTIDOffset;
    bool fanu;
    std::vector<TString> fanuFileVector;

  };
  

  //____________________________________________________________________________
  TGMuon::TGMuon(fhicl::ParameterSet const& pset): fTIDOffset(100),fanu(true)
  {
    this->reconfigure(pset);
    produces< std::vector<simb::MCTruth> >();
    produces< sumdata::RunData, art::InRun >();
    produces< sumdata::POTSummary, art::InSubRun >();
    fRand = new TRandom3();
    fRand->SetSeed();
    totalpot=0.;    
    fanu       = pset.get< bool  >("AntiNeutrinos", false);
  }

  //____________________________________________________________________________
  TGMuon::~TGMuon()
  {
  if(fRand) delete fRand;
  }

  //____________________________________________________________________________

  //____________________________________________________________________________
  void TGMuon::beginRun(art::Run& run)
  {
    // grab the geometry object to see what geometry we are using
    art::ServiceHandle<geo::Geometry> geo;
    //geo::DetId_t detid = geo->DetId();
    std::unique_ptr<sumdata::RunData> runcol(new sumdata::RunData("argoneut"));
    run.put(std::move(runcol));
    if (!fanu) return;
    // Now, find the antineutrino files.


    //    TString anuFiles("/argoneut/data/rootfiles_ART_MINOS/AntiNeutrino");
    TString anuFiles("/argoneut/data/minos/rhc");
    //    const char* ordir(TSystem().pwd() );

    FILE *fp = popen("pwd", "r");
    while (fgets(fBuffer, sizeof(fBuffer), fp) != NULL)
      {
	std::cout << "TGMuon: Output from pwd is: " << fBuffer << '\n';
	std::cout << "TGMuon: sizeof(fBuffer) is: " << sizeof(fBuffer) << '\n';
      }
    pclose(fp);

    fSaveDir = gDirectory;
    TString fpwd(gSystem->pwd());


    TList* l = TSystemDirectory("",anuFiles).GetListOfFiles();
    Int_t length = l->GetSize();
    Int_t ii(0);

    while ( ii<length )
      {
	TSystemFile* currentFile = (TSystemFile*)(l->At(ii));
	TString rootFile(currentFile->GetName());
	std::cout << " currentFile is " << rootFile << std::endl;
	fanuFileVector.push_back(anuFiles+"/"+rootFile);
	ii++;
      }

    return;

  }

  void TGMuon::endSubRun(art::SubRun& sr)
  {
    std::unique_ptr<sumdata::POTSummary> p(new sumdata::POTSummary());    
    p->totpot = totalpot;
    p->totgoodpot = totalpot;    
    sr.put(std::move(p));    
    return;
  }

  //____________________________________________________________________________
  void TGMuon::produce(art::Event& evt)
  {
    ///unique_ptr allows ownership to be transferred to the art::Event after the put statement
    std::unique_ptr< std::vector<simb::MCTruth> > truthcol(new std::vector<simb::MCTruth>);
    simb::MCTruth truth;
    truth.SetOrigin(simb::kSingleParticle);
    Sample(truth);
    truthcol->push_back(truth);
    evt.put(std::move(truthcol));

    return;
  }

  //____________________________________________________________________________
  void TGMuon::Sample(simb::MCTruth &mct) 
  {
      //loop through MINOS root files	  
      std::string file = "/argoneut/data/outstage/spitz7/allminos_neutrinodata/allminos_neutrinodata.root";        

      if (fanu)
	{
	  int anuIndex = (int)fRand->Uniform(1,fanuFileVector.size());
	  while (anuIndex < 2)  // first 2 are . and ..
	    {
	      anuIndex = (int)fRand->Uniform(1,fanuFileVector.size());
	    }
	  if (fanuFileVector.size()) file = (std::string)(fanuFileVector.at(anuIndex));
	}

      int snarl=0;
      int run=0;
      int subRun=0;
      int flag=0;
      double r=0;
      double x_offset=117.4; 
      double y_offset=19.3;
      double z_offset=150.; //force particle to be well upstream of ArgoNeuT
  
            TFile *f = new TFile(file.c_str());
            TTree *minitree = (TTree*)f->Get("minitree");
            minitree->SetBranchAddress("run",&frun);  
            minitree->SetBranchAddress("subRun",&fsubRun);            
            minitree->SetBranchAddress("trkIndex",&ftrkIndex);
            minitree->SetBranchAddress("ntrkstp",&fntrkstp);
            minitree->SetBranchAddress("snarl",&fsnarl);
            minitree->SetBranchAddress("trtgtd",&ftrtgtd);
            minitree->SetBranchAddress("trkVtxX",&ftrkVtxX);
            minitree->SetBranchAddress("trkVtxY",&ftrkVtxY);
            minitree->SetBranchAddress("trkVtxZ",&ftrkVtxZ);
            minitree->SetBranchAddress("trkdcosx",&ftrkdcosx);
            minitree->SetBranchAddress("trkdcosy",&ftrkdcosy);
            minitree->SetBranchAddress("trkdcosz",&ftrkdcosz);
            minitree->SetBranchAddress("trkmom",&ftrkmom);
            minitree->SetBranchAddress("trkqp",&ftrkqp);            
            Long64_t nentries = minitree->GetEntries();
            //Long64_t nbytes = 0;           
      
      int flag2=0;
      while(flag2==0)    
      {      
            r=(int)fRand->Uniform(1,nentries-1);
            minitree->GetEntry(r);
            run=frun;
            subRun=fsubRun;
            snarl=fsnarl;
            //count POT even if the particle is not passed to Geant
            totalpot+=ftrtgtd*pow(10,12);
          
      for(int i=r-ftrkIndex;i<nentries;i++)
      {
          minitree->GetEntry(i);
           
          if(frun==run && fsubRun==subRun && fsnarl==snarl)
          flag=1;           
           
    	  if(flag==1 && fsnarl!=snarl)
     	  break;
        
     	 TVector3 x;
	 x[0]=100.*ftrkVtxX - x_offset;
    	 x[1]=100.*ftrkVtxY + y_offset;
    	 x[2]=100.*ftrkVtxZ - z_offset;        
        //don't bother passing the particle to geant if it's way outside the argoneut detector window   
      	if(100.0*ftrkVtxZ > 20. || fabs(x[0]+23.5)>100. || fabs(x[1])>100.)
      	continue;

      	// Choose momentum
      	double p = 0.0;
      	double m = 0.0;
      	p=ftrkmom;
      
      	if(!p)
      	continue;
      
      	int pdg=0;
      	
      	if(ftrkqp>0)
      	pdg=-13;
      	else
     	pdg=13;
      
      	TParticlePDG* pdgp = TDatabasePDG::Instance()->GetParticle(pdg);
      	if (pdgp) m = pdgp->Mass();

      	TLorentzVector pos(x[0], x[1], x[2], 0.0);
      	TLorentzVector pvec(ftrkdcosx*p,ftrkdcosy*p,ftrkdcosz*p,sqrt(p*p+m*m));
    
      	int trackid = -1*(i+1+fTIDOffset); // set track id to -i as these are all primary particles and have id <= 0
      	std::string primary("primary");

      	simb::MCParticle part(trackid, pdg, primary);
      	part.AddTrajectoryPoint(pos, pvec);       
      	mct.Add(part);
	flag2=1;      
      }//for loop
      }//while flag2 loop
       minitree->Delete();
       f->Close();
       f->Delete();

   return;
  }

  DEFINE_ART_MODULE(TGMuon)

}//end namespace t962
////////////////////////////////////////////////////////////////////////
